var fs     = require('fs'),
	fse	   = require('fs-extra'),
    mkdirp = require('mkdirp'),
    _      = require('lodash'),
    path   = require('path'),
    async  = require('async'),
	CircularJSON = require('circular-json'),
	fExtra = require('fs-extra'),
	_ = require('lodash');

require('string.prototype.startswith');

var UNDEFINED, exportObject = exports, reportDate;

function sanitizeFileName(name){
    return sanitizeName(name);
}

function sanitizeName(name){
    name = name.replace(/\s+/gi, '-'); // Replace white space with dash
    return name.replace(/[^a-zA-Z0-9\-]/gi, ''); // Strip any special character
}

function trim(str) { return str.replace(/^\s+/, "" ).replace(/\s+$/, "" ); }

function elapsed(start, end) { return (end - start)/1000; }

function elapsed1(start, end) { 
	return (Date.parse(end) - Date.parse(start))/1000;
}

function isFailed(obj) { return obj.status === "failed"; }

function isSkipped(obj) { return (obj.status === 'pending' || obj.status === 'skipped'); }

function isDisabled(obj) { return obj.status === "disabled"; }

function writeResults(suite, savePath) {
		var fs = require("fs");
		var path = require("path");
		var filePath = savePath + '/results';
		require("mkdirp").sync(filePath);		
		console.log('jsonFile', path.join(filePath, sanitizeName(suite.description)));
		var jsonFile = fs.openSync(path.join(filePath, sanitizeName(suite.description)) + '.json', "w");	
		fs.writeSync(jsonFile, CircularJSON.stringify(suite), 0);
		fs.closeSync(jsonFile);
		return;
}

function isPassed(obj) { return obj.status === "passed"; }

function isArray(what) {
    return Object.prototype.toString.call(what) === '[object Array]';
}

function getPath(filename, dir) {
  if (dir) {
    fExtra.ensureDirSync(dir);
    return path.join(dir, filename);
  } else {
    return path.join(__dirname, filename);
  }
}

function readFile(filename) {
  return fs.readFileSync(filename, 'utf-8');
}

function deepCopy(source) {
	return JSON.parse(JSON.stringify(source));
}

function extend(tgt, src) {
    for (var prop in src) {
        if (src.hasOwnProperty(prop)) {
            tgt[prop] = src[prop];
        }
    }
    return tgt;
}


function log(str) {
    var con = global.console || console;
    if (con && con.log) {
        con.log(str);
    }
}

function rmdir(dir) {
    try {
        var list = fs.readdirSync(dir);
        for (var i = 0; i < list.length; i++) {
            var filename = path.join(dir, list[i]);
            var stat = fs.statSync(filename);

            if (stat.isDirectory()) {
                // rmdir recursively
                rmdir(filename);
            } else {
                // rm fiilename
                fs.unlinkSync(filename);
            }
        }
        fs.rmdirSync(dir);
    }catch (e) { log("problem trying to remove a folder:" + dir); }
}


/* 
	This function is responsible for generating the test results and it will be placed in results folder. 
	This results will be used for generating the consolidated report.
*/

function TestResultsGenerator(options) {

    var self = this;

    self.started = false;
    self.finished = false;
	self.screenshotsDir = 'screenshots';
	
    // sanitize arguments
    options = options || {};
    self.takeScreenshots = options.takeScreenshots === UNDEFINED ? true : options.takeScreenshots;
    self.reportDir = options.reportDir || '';
    self.takeScreenshotsOnlyOnFailures = options.takeScreenshotsOnlyOnFailures === UNDEFINED ? false : options.takeScreenshotsOnlyOnFailures;
	
    self.cleanReportDir = options.cleanReportDir === UNDEFINED ? true : options.cleanReportDir;

    var suites = [],
        currentSuite = null,
        totalSpecsExecuted = 0,
        totalSpecsDefined,
		
        // when use use fit, jasmine never calls suiteStarted / suiteDone, so make a fake one to use
        fakeFocusedSuite = {
            id: 'focused',
            description: 'focused specs',
            fullName: 'focused specs'
        };
	
    var __suites = {}, __specs = {};
	
    function getSuite(suite) {
        __suites[suite.id] = extend(__suites[suite.id] || {}, suite);
        return __suites[suite.id];
    }
	
    function getSpec(spec) {
        __specs[spec.id] = extend(__specs[spec.id] || {}, spec);
        return __specs[spec.id];
    }
	
	/* Method is invoked when jasmine gets started */
    self.jasmineStarted = function(summary) {
        totalSpecsDefined = summary && summary.totalSpecsDefined || NaN;
        exportObject.startTime = new Date();
        self.started = true;

        if (self.cleanReportDir)
            rmdir(self.reportDir);

    };

	/* Method is invoked when suite execution gets started */
    self.suiteStarted = function(suite) {
        suite = getSuite(suite);
        suite._startTime = new Date();
        suite._specs = [];
        suite._suites = [];
        suite._failures = 0;
        suite._skipped = 0;
        suite._disabled = 0;
        suite._parent = currentSuite;
        if (!currentSuite) {
            suites.push(suite);
        } else {
            currentSuite._suites.push(suite);
        }
        currentSuite = suite;
    };

	/* Method is invoked when spec execution gets started */
    self.specStarted = function(spec) {
        if (!currentSuite) {
            // focused spec (fit) -- suiteStarted was never called
            self.suiteStarted(fakeFocusedSuite);
        }
        spec = getSpec(spec);
        spec._startTime = new Date();
        spec._suite = currentSuite;
        currentSuite._specs.push(spec);
    };

	/* Method is invoked when spec execution is finished */
    self.specDone = function(spec) {
        spec = getSpec(spec);
        spec._endTime = new Date();
        if (isSkipped(spec)) { spec._suite._skipped++; }
        if (isDisabled(spec)) { spec._suite._disabled++; }
        if (isFailed(spec)) { spec._suite._failures++; }
        totalSpecsExecuted++;

        if ((self.takeScreenshots && !self.takeScreenshotsOnlyOnFailures) ||
            (self.takeScreenshots && self.takeScreenshotsOnlyOnFailures && isFailed(spec))) {
                spec.screenshot = sanitizeName(spec.description) + '.png';

            browser.takeScreenshot().then(function (png) {
                var screenshotPath = path.join(
                    self.reportDir,
                    self.screenshotsDir,
                    spec.screenshot
                );

                mkdirp(path.dirname(screenshotPath), function (err) {
                    if (err) {
                        throw new Error('Could not create directory for ' + screenshotPath);
                    }
                    writeScreenshot(png, screenshotPath);
                });
            });
        }


    };
	
	/* Method is invoked when suite execution is finished */
    self.suiteDone = function(suite) {
        suite = getSuite(suite);
        if (suite._parent === UNDEFINED) {
            // disabled suite (xdescribe) -- suiteStarted was never called
            self.suiteStarted(suite);
        }
        suite._endTime = new Date();
        currentSuite = suite._parent;
    };

	/* Method is invoked when jasmine is done with the execution */
    self.jasmineDone = function() {
        if (currentSuite) {
            // focused spec (fit) -- suiteDone was never called
            self.suiteDone(fakeFocusedSuite);
        }
		
		for (var i = 0; i < suites.length; i++) {
            if(suites[i]._parent === null) {
				log('writing json ');
				writeResults(suites[i], self.reportDir);
				log('wrote json ');
				break;
			}
        }
        self.finished = true;
        exportObject.endTime = new Date();
    };

	 var writeScreenshot = function (data, filename) {
		var stream = fs.createWriteStream(filename);
		stream.write(new Buffer(data, 'base64'));
		stream.end();
	};

    return this;
}

/* 
	This function is responsible for creating the html report based on the templates using the results.
	It takes baseReportDir & reportDir inputs
*/

function TestReportGenerator(baseReportDir, reportDir, reportDirSource, title, deployPath) {
	
	this.baseReportDir = baseReportDir;
	this.reportDir = reportDir;
	this.reportDirSource = reportDirSource;
	this.title = title;
	this.deployPath = deployPath;
		
	/* This object holds the report summary such as test, suite and spec statuses */

	var reportSummary = {
		title: this.title,
		suites: 0,
		specs: 0,
		runs: 0,
		failedSuites: 0,
		passedSuites: 0,
		failedSpecs: 0,
		passedSpecs: 0,
		skippedSpecs: 0,
		disabledSpecs: 0,
		elapsed: 0
	};
	
	/* This object holds the whole test run summary */
	var testRunSummary = {
		title: this.title,
		runs: 0,
		passed: 0,
		failed: 0,
		elapsed: 0
	};

	/* This object holds each suite summary */
	var suiteSummaryObj = {
		id: '',
		suites: 0,
		runs: 0,
		passedSuites: 0,
		failedSuites: 0,
		specs: 0,
		failed: 0,
		skipped: 0,
		passed: 0,
		disabled: 0,
		elapsed: 0
	};

	/* This object holds each test summary */
	var testSummaryObj = {
		id: '',
		runs: 0,
		passed: 0,
		failed: 0,
		elapsed:0,
		min: 0,
		max: 0,
		avg: 0
	};

	/* This object holds each spec details */
	var testSpecObj = {	
		description: '',
		suite: '',
		status: '',
		screenshot: '',
		messages: [],
		stacks: [],
		elapsed: 0,
		data: ''
	};

	/* This object holds each suite details */
	var testSuiteObj = {
		title: '',
		testSpecs: [],
		elapsed: 0,
		failedSpecs: 0,
		passedSpecs:0
	};
	
	/* This object holds test case summary */
	var testCaseSummaryObj = {
		passed: 0,
		failed: 0,
		total: 0,
		elapsed: 0
	};
	
	/* This object holds each scenario details such as test case summary and case details */
	var suiteDetailObj = {
		description: '',
		testCaseSummaries: [],
		testCaseDetails: []
	};
	
	/* This object holds each test case details */
	var testCaseDetailObj = {
		data: {},
		testSuites: [],
		status: ''
	};
		

	var suiteDetails = {};
	 
	var suiteSummaries = {};

	var testSummaries = {};

	var suiteResults = [];
	
	/* updates the spec statuses in summary */
	function updateSpecStatus(suiteSummary, spec) {
		if(isFailed(spec)) {
			suiteSummary.failed++;
			reportSummary.failedSpecs++;
		} else if(isSkipped(spec.status)) {
			suiteSummary.skipped++;
			reportSummary.skippedSpec++;
		} else if(isDisabled(spec)) {
			suiteSummary.disabled++;
			reportSummary.disabledSpecs++;
		} else {
			suiteSummary.passed++;
			reportSummary.passedSpecs++;
		}
	}

	/* checks whether the suite is success */
	function isSuiteSuccess(suite) {
		return !(suite._failures > 0);
	}

	/* updates the elapsed time in report summary with max time */
	function updateReportElapsed(time) {
		if(reportSummary.elapsed < time) {
			reportSummary.elapsed = time;
		}
	}
	
	/* updates the elapsed time in test run summary with max time */
	function updateTestRunElapsed(time) {
		if(testRunSummary.elapsed < time) {
			testRunSummary.elapsed = time;
		}
	}

	/* updates the number of test runs in all summary objects */
	function updateRuns(suiteSummary, testSummary) {
		suiteSummary.runs++;
		testSummary.runs++;
		testRunSummary.runs++;
		reportSummary.runs++;
	}

	/* updates the number of passed/failed suites in all summary objects */
	function updateSuiteStatus(suite, suiteSummary) {
		if(isSuiteSuccess(suite)) {
			reportSummary.passedSuites++;
			suiteSummary.passedSuites++;
		} else {
			reportSummary.failedSuites++;
			suiteSummary.failedSuites++;
		}
	}

	/* updates the number of suites and statuses in summary objects */
	function updateSuiteSummary(suite, suiteSummary) {
		reportSummary.suites++;
		suiteSummary.suites++;
		updateSuiteStatus(suite, suiteSummary);	
	}

	/* updates the number of passed/failed specs in suite & report summary objects */
	function updateSpecStatus(spec, suiteSummary) {
		if(isFailed(spec)) {
			suiteSummary.failed++;
			reportSummary.failedSpecs++;
		} else if(isSkipped(spec)) {
			suiteSummary.skipped++;
			reportSummary.skippedSpecs++;
		} else if(isDisabled(spec)) {
			suiteSummary.disabled++;
			reportSummary.disabledSpecs++;
		} else {
			suiteSummary.passed++;
			reportSummary.passedSpecs++;
		}
	}
	
	/* updates the test timings (min/max/avg) in test summary */
	function updateTestTimings(testSummary, elapsed) {
		if(testSummary.min === 0 && testSummary.max === 0 && testSummary.avg === 0) {
			testSummary.min = elapsed;
			testSummary.max = elapsed;
			testSummary.avg = elapsed;
			return;
		}
		if(testSummary.min > elapsed) {
			testSummary.min = elapsed;
		}
		if(testSummary.max < elapsed) {
			testSummary.max = elapsed;
		}
		testSummary.avg = testSummary.elapsed/testSummary.runs;
	}

	/* updates the number of passed/failed tests */
	function updateTestSummary(testSummary, failures) {
		if(failures > 0) {
			testSummary.failed++;
			testRunSummary.failed++;
		} else {
			testSummary.passed++;
			testRunSummary.passed++;
		}		
	}
	
	/* updates the number of specs in summary objects */
	function updateSpecSummary(suite, suiteSummary) {
		reportSummary.specs += suite._specs.length;
		suiteSummary.specs += suite._specs.length;
	}

	/* updates the elapsed time in all summary objects */
	function updateElapsed(suite, suiteSummary, testSummary) {
		suiteSummary.elapsed =  elapsed1(suite._startTime, suite._endTime);
		testSummary.elapsed = suiteSummary.elapsed;
		updateReportElapsed(suiteSummary.elapsed);
		updateTestRunElapsed(testSummary.elapsed);
	}

	/* generates the summaries for all the scenario */
	function generateSummaries() {
		suiteResults.forEach( function(suiteResult) {
			var id = sanitizeName(suiteResult.description);
			
			if(suiteResult._suites) {
				var suiteSummary = deepCopy(suiteSummaryObj);
				var testSummary = deepCopy(testSummaryObj);
				var failures = 0;
				
				suiteSummary.id = id;
				testSummary.id = id;
				updateElapsed(suiteResult, suiteSummary, testSummary); //updating the total time taken for all the test runs
				
				suiteResult._suites.forEach(function(driveSuite) {				
					updateRuns(suiteSummary, testSummary); //updating the test runs
					
					generateDriveSummary(driveSuite, suiteSummary); //generating the each test run summary
					updateTestTimings(testSummary, elapsed1(driveSuite._startTime, driveSuite._endTime));
					updateTestSummary(testSummary, (suiteSummary.failedSuites - failures));						
					failures = suiteSummary.failedSuites;
				});
				suiteSummaries[suiteResult.description] = suiteSummary;
				testSummaries[suiteResult.description] = testSummary;
			}
		});
	}

	/* Generates the summary for the driven suite */
	function generateDriveSummary(driveSuite, suiteSummary) {	
		if(driveSuite._suites) {
			driveSuite._suites.forEach(function(suite) {
				generateSummary(suite, suiteSummary);
			});
		}
	}
		
	/* generates the summary for each suite */
	function generateSummary(suite, suiteSummary) {
		updateSuiteSummary(suite, suiteSummary);	
		
		if(suite._specs) {
			updateSpecSummary(suite, suiteSummary);
			suite._specs.forEach(function(spec) {
				updateSpecStatus(spec, suiteSummary);
				if(isArray(spec._suites) && spec._suites.length > 0) {
					spec._suites.forEach(function(specSuite) {
						generateSummary(specSuite, suiteSummary);
					});
				}
			});
		}
		if(suite._suites) {
			suite._suites.forEach(function(subSuite) {
				generateSummary(subSuite, suiteSummary);
			});
		}
	}
	
		
	/* gets the results of all scenario ran from the results directory */
	function getSuiteResults(resultsDir) {
		var jsonFiles = fs.readdirSync(resultsDir)
		.map(function(fileName) { 
			return { name: fileName,
				time: fs.statSync(resultsDir + '/' + fileName).mtime.getTime(),
				path: path.join(resultsDir, fileName)
			}; 
		})
		.sort(function(file1, file2) { return file1.time - file2.time; })
		.map(function(file) { return file.path; });
		
		jsonFiles.forEach(function(file) {
			console.log('File:', file);
			suiteResults.push(require(file));
		});
	}
	
	
	/* generates all the details of each suites */
	function generateSuiteDetails() {
		suiteResults.forEach( function(suiteResult) {//scenario suite
			suiteDetails[suiteResult.description] = generateSuiteDetail(suiteResult);
		});
	}
	
	/* generate details for each scenario result */
	function generateSuiteDetail(suiteResult) {		
		var suiteDetail = deepCopy(suiteDetailObj);
		suiteDetail.description = suiteResult.description;
		
		suiteResult._suites.forEach(function(drivenSuite) { //drive suite
			var testCaseDetail = generateTestCaseDetail(drivenSuite);
			suiteDetail.testCaseDetails.push(testCaseDetail);
			suiteDetail.testCaseSummaries.push(generateTestCaseSummary(testCaseDetail, drivenSuite));
		});
		return suiteDetail;
	}
	
	/* generates test case details for each driven suite */
	function generateTestCaseDetail(drivenSuite) {
		var testCaseDetail = deepCopy(testCaseDetailObj);
		testCaseDetail.data = generateTestRunData(drivenSuite.data);
		
		if(drivenSuite._suites) {//running suites		
			drivenSuite._suites.forEach(function (suite) {// running suite
				var suites = [];
				generateTestSuites(suite, suites);
				suites.forEach(function (suite1) {// pushing the test suites one by one
					testCaseDetail.testSuites.push(suite1);
				});
			});
		}
		return testCaseDetail;
	}
	
	/* generates test case summary for each driven suite */
	function generateTestCaseSummary(testCaseDetail, drivenSuite) {
		var testCaseSummary = deepCopy(testCaseSummaryObj);
		
		testCaseDetail.testSuites.forEach(function(suite) {
			suite.testSpecs.forEach(function(spec) {
				if(isFailed(spec)) { testCaseSummary.failed++; };
				if(isPassed(spec)) { testCaseSummary.passed++; };
				testCaseSummary.total++;
			});
		});
		testCaseSummary.elapsed = elapsed1(drivenSuite._startTime, drivenSuite._endTime);
		testCaseDetail.status = ((testCaseSummary.failed > testCaseSummary.passed)? 'failed' : (testCaseSummary.passed > 0 && testCaseSummary.failed == 0) ? 'passed' : '');
		
		return testCaseSummary;
	}

	/* generates data thats being used for spec */
	function generateSpecData(testData) {	
		var data = '|';
		if(testData) {
			for(var prop in testData) {
				data += prop.toUpperCase() + ': ' + testData[prop] + ' | ';
			}
		}
		return data;
	}
	
	/* generates formatted the test data used */
	function generateTestRunData(testData) {
		var dataTh = '<thead class="text-primary">';
		var dataRw = '<tbody><tr>';
		var data = '';
		for(var prop in testData) {
			dataTh += '<th>' + prop.toUpperCase() + '</th>';
			dataRw += '<td>' + testData[prop] + '</td>';
		}
		dataTh += '</thead>';
		dataRw += '</tr></tbody>';
		return (dataTh + dataRw);
	}
	
	/* generates test suites details */
	function generateTestSuites(suite, testSuites) {
		var testSuite = generateTestSuite(suite);
		testSuites.push(testSuite);
		if(suite._suites) {
			suite._suites.forEach(function(subSuite) {
				generateTestSuites(subSuite, testSuites);
			});
		}
	}

	/* generates test suite details of each suite */
	function generateTestSuite(suite) {
		var testSt = deepCopy(testSuiteObj);
		testSt.title = suite.description;
		testSt.elapsed = elapsed1(suite._startTime, suite._endTime);
		
		if(suite._specs) {
			suite._specs.forEach(function (spec) {
				var testSpec = generateTestSpec(spec);
				if(isFailed(testSpec)) {testSt.failedSpecs++; };
				if(isPassed(testSpec)) {testSt.passedSpecs++; };
				testSt.testSpecs.push(testSpec);
			});
		}
		return testSt;
	}

	/* generates each spec details */
	function generateTestSpec(spec) {	
		var specDet = deepCopy(testSpecObj);
		specDet.description = spec.description;
		specDet.status = spec.status;
		specDet.screenshot = spec.screenshot;
		specDet.elapsed = elapsed1(spec._startTime, spec._endTime);
		specDet.data = generateSpecData(spec.data);
		populateStacksMessages(spec, specDet);
		return specDet;	
	}

	/* populates stack traces and messages in spec */
	function populateStacksMessages(spec, specDet) {
		spec.failedExpectations.forEach(function (failedExpectation) {
			specDet.messages.push(failedExpectation.message);
			specDet.stacks.push(failedExpectation.stack);
		});
	}
	
	/* generates the html report */
	function generateHTMLReport(baseReportDir, reportDir) {
		generate(baseReportDir, reportDir);
	}

	/* generates the report based on the templates and keeps in reportDir */
	function generate(baseReportDir, reportDir) {
		console.log('reportSummary:', reportSummary);
		console.log('testSummaries:', testSummaries);
		console.log('suiteSummaries:', suiteSummaries);
		console.log('suiteDetails:', suiteDetails);	
		
		console.log('baseReportDirrrrrr:', baseReportDir);
		console.log('reportDirrrrrr:', reportDir);
		
		fExtra.renameSync(baseReportDir, reportDir);
		fExtra.writeFileSync(
			getPath('report.html', reportDir),
		  _.template(readFile(getPath('./templates/report.tmpl')))({
				styles: readFile(getPath('./assets/report.css')),
				reportSummary: _.template(readFile(getPath('./templates/reportsummary.tmpl')))({
					reportSummary: reportSummary,
					testRunSummary: testRunSummary,
					_: _,
				}),
				suiteSummaries: suiteSummaries,
				testsSummary: _.template(readFile(getPath('./templates/suitessummary.tmpl')))({
					testSummaries: testSummaries,
					_: _,
				}),
				chartScript: readFile(getPath('./assets/chartist.min.js')),
				charts: _.template(readFile(getPath('./templates/charts.tmpl')))({
					reportSummary: reportSummary,
					suiteSummaries: suiteSummaries,
					testRunSummary: testRunSummary,
					_: _
				}),
				suiteDetails: _.template(readFile(getPath('./templates/suitedetails.tmpl')))({
					suiteSummaries: suiteSummaries,
					_: _,
					suiteDetails: suiteDetails
				})
		  })
		);
	}
	
	function copyToDeployFolder(reportDirSource,deployPath){
		fse.copySync(reportDirSource, deployPath)
	}
	
	/* generates the test report using the results */
	this.generateReport = function() {
		console.log('baseReportDir:', this.baseReportDir);
		getSuiteResults(getPath('results', this.baseReportDir));
		generateSummaries();		
		generateSuiteDetails();
		generateHTMLReport(this.baseReportDir, this.reportDir);
		copyToDeployFolder(this.reportDirSource,this.deployPath);
	}
}

/* exports both functions of generating the test results and reports respectivley */
module.exports = {TestResultsGenerator, TestReportGenerator};
